Require Import Program Vtac hpattern.
Require Import FunctionalExtensionality Compare_dec Le Image List Ensembles Mult.
Require Import Classical IndefiniteDescription ChoiceFacts.
Require Import lang.

Set Implicit Arguments.
Unset Strict Implicit.
Open Scope string_scope.

(*----------------------------------------------------------------------------*)
(** ** Classical Logic *)
(*----------------------------------------------------------------------------*)

(** We assume classical logic, functional and propositional extensionality. *)

Lemma prop_ext : ClassicalFacts.prop_extensionality.
Proof.
red; intros; exploit (@Extensionality_Ensembles () (fun _ => A) (fun _ => B)).
  by split; red; intros; unfold In in *; intuition.
by intros EQ; change A with ((fun _ => A) ()); change B with ((fun _ => B) ()); rewrite EQ.
Qed.

Lemma proof_irr: ClassicalFacts.proof_irrelevance.
Proof. exact (ClassicalFacts.ext_prop_dep_proof_irrel_cic prop_ext). Qed.

Lemma proj1_sig_inj: forall A (N : A -> Prop) (a a': sig N) (EQ: `a = `a'), a = a'.
Proof.
intros; destruct a, a'; simpls; subst; f_equal; eauto using proof_irr.
Qed.

Lemma sig_eta: forall A (P: A -> Prop) (x : sig P) pf, exist P (`x) pf = x.
Proof. by intros; destruct x; simpls; apply proj1_sig_inj. Qed.

Lemma LEM: forall (P: Prop), {P} + {~P}.
Proof.
intros; eapply constructive_definite_descr_excluded_middle; eauto using classic.
repeat intro; eapply constructive_indefinite_description in H.
by destruct H as (a & H); destruct H; eauto.
Qed.

Lemma or2sum: forall P Q (OR: P \/ Q), {P} + {Q}.
Proof.
intros; destruct (LEM P); [by left|]; destruct (LEM Q); [by right|].
by exfalso; intuition.
Qed.

Ltac gen_cid :=
  match goal with [|- context[@constructive_indefinite_description ?A ?B ?C]] =>
    generalize (@constructive_indefinite_description A B C)
  end.

Ltac gen_eqrefl :=
  match goal with [|-context[@eq_refl ?A ?B]] => 
    generalize (@eq_refl A B) 
  end.

(*----------------------------------------------------------------------------*)
(** ** Countably infinite sets *)
(*----------------------------------------------------------------------------*)

Definition surjective A B (f: A -> B) : Prop := forall b, exists a, b = f a.

Definition countably_infinite_set (A: Type) :=
  exists f: A -> nat, exists g: nat -> A,
    (forall a, g (f a) = a) /\
    (forall n, f (g n) = n).

Definition countably_infinite_set2 (A:Type) :=
  exists f: A -> nat, exists g: nat -> A,
    injective _ _ f /\ injective _ _ g.

Definition countably_infinite A (N: A -> Prop) :=
  countably_infinite_set (sig N).

Definition ndisj A (N1 N2: A -> Prop) := forall n, N1 n -> N2 n -> False.

Definition nunion A (N1 N2: A -> Prop) := fun n => N1 n \/ N2 n. 

Lemma ndisj_mon : 
  forall 
    A N1' N2' (DIS: ndisj (A:=A) N1' N2')
    (N1: A->Prop) (LT1: forall a (EL: N1 a), N1' a)
    (N2: A->Prop) (LT2: forall a (EL: N2 a), N2' a),
  ndisj N1 N2.
Proof. red; intros; eauto. Qed.

Lemma even_or_odd: forall n, exists i, n = 2 * i \/ n = S (2 * i).
Proof. by intros; induction n; des; subst; eauto; exists (S i); omega. Qed.

(** A countably infinite set can be divided into two disjoint countably 
    infinite sets. *)

Lemma countably_infinite_split: 
  forall A (N: A -> Prop) (INF: countably_infinite N),
    exists N1 N2, 
      <<Nci1: countably_infinite N1>> /\ <<Nci2: countably_infinite N2>> /\
      <<Nunion: N = nunion N1 N2>> /\ <<Ndisj: ndisj N1 N2>>.
Proof.
  intros; do 2 red in INF; des.
  set (N1 a := exists i, a = proj1_sig (g (2 * i))).
  set (N2 a := exists i, a = proj1_sig (g (S(2 * i)))).
  set (f1 (a : sig N1) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (f2 (a : sig N2) :=
    projT1 (constructive_indefinite_description _ (proj2_sig a))).
  set (g1 i := exist N1 (proj1_sig (g (2 * i))) (ex_intro _ _ eq_refl)).
  set (g2 i := exist N2 (proj1_sig (g (S(2 * i)))) (ex_intro _ _ eq_refl)).

  assert (F1F: forall a Pa1 Pa, f (exist _ a Pa) = 2 * (f1 (exist _ a Pa1))).
    intros; unfold f1; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.
  assert (F2F: forall a Pa1 Pa, f (exist _ a Pa) = S(2 * (f2 (exist _ a Pa1)))).
    intros; unfold f2; gen_cid; intros (n & Pn); simpls.
    by rewrite <- INF0; f_equal; subst; eapply proj1_sig_inj.

  exists N1 N2; split.
    exists f1 g1; split; intros. 
      destruct a as (a & Pa); unfold g1. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(2*i)))).
      gen_eqrefl; hrewrite <-(F1F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.
    unfold f1, g1; gen_cid; intros (n' & Pn'); simpls.
    apply proj1_sig_inj in Pn'.
    cut (n+(n+0)=n'+(n'+0)); [by omega|].
    by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  split.
    exists f2 g2; split; intros.
      destruct a as (a & Pa); unfold g2. 
      assert (Pa' : N a) by (rdes Pa; subst; apply (proj2_sig (g(S(2*i))))).
      gen_eqrefl; hrewrite <-(F2F _ Pa Pa'); rewrite INF.
      by simpls; eauto using proj1_sig_inj.      
    unfold f2, g2; gen_cid; intros (n' & Pn'); simpls.
    apply proj1_sig_inj in Pn'.
    cut (S(n+(n+0))=S(n'+(n'+0))); [by omega|].
    by rewrite <-INF0; rewrite <-INF0 at 1; rewrite Pn'.
  split.
  - red; extensionality a; apply prop_ext; unfold nunion, N1, N2; split; intros X.
    + assert (Y := X); change a with (proj1_sig(exist _ _ X)) in Y.
      rewrite <- (INF (exist _ _ X)) in Y.
      destruct (even_or_odd (f (exist _ _ X))); des; [left|right];
       by exists x; rewrite <- H; rewrite INF.
    + by des; subst; apply proj2_sig.
  - intros n (i & X) (j & Y); subst. 
    eapply proj1_sig_inj in Y.
    cut (2*i = S(2*j)); [by omega|].
    by rewrite <- INF0; rewrite <- INF0 at 1; rewrite Y.
Qed.

